import logging
import asyncio
import argparse

import i2plib.sam 
import i2plib.aiosam
import i2plib.utils

BUFFER_SIZE = 65536

class I2PTunnel(object):
    """I2P Tunnel object

    This object is returned by tunnel creation coroutines

    :param name: Tunnel session name
    :param session_writer: Session socket writer instance
    :param future: Tunnel future
    """

    def __init__(self, name, session_writer, future):
        self.name = name
        self.session_writer = session_writer
        self.future = future

    def stop(self):
        """Stop the tunnel"""
        self.session_writer.close()
        self.future.cancel()

async def proxy_data(reader, writer):
    """Proxy data from reader to writer"""
    try:
        while True:
            data = await reader.read(BUFFER_SIZE)
            if not data:
                break
            writer.write(data)
    except Exception as e:
        logging.debug('proxy_data_task exception {}'.format(e))
    finally:
        try:
            writer.close()
        except RuntimeError:
            pass
        logging.debug('close connection')

async def client_tunnel(local_address, remote_destination, loop=None, 
                            private_key=None, session_name=None,
                            sam_address=i2plib.sam.DEFAULT_ADDRESS):
    """Run a client tunnel in the event loop.

    If you run a client tunnel with a local address ("127.0.0.1", 6668) and
    a remote destination "irc.echelon.i2p", all connections to 127.0.0.1:6668 
    will be proxied to irc.echelon.i2p.

    :param local_address: A local address to bind a remote destination to. E.g.
                        ("127.0.0.1", 6668)
    :param remote_destination: Remote I2P destination, can be either .i2p 
                        domain, .b32.i2p address, base64 destination or 
                        i2plib.Destination instance
    :param session_name: (optional) Session nick name. A new session nickname is
                        generated if not specified.
    :param private_key: (optional) Private key to use in this session. Can be 
                        a base64 encoded string, i2plib.sam.PrivateKey instance
                        or None. TRANSIENT destination is used when it is None.
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :return: an instance of i2plib.tunnel.I2PTunnel
    """
    session_name = session_name or i2plib.sam.generate_session_id()
    reader, writer = await i2plib.aiosam.create_session(session_name,
                style="STREAM", sam_address=sam_address, loop=loop,
                private_key=private_key)

    async def handle_client(client_reader, client_writer):
        """Handle local client connection"""
        remote_reader, remote_writer = await i2plib.aiosam.stream_connect(
                session_name, remote_destination, sam_address=sam_address,
                loop=loop)
        asyncio.ensure_future(proxy_data(remote_reader, client_writer), 
                              loop=loop)
        asyncio.ensure_future(proxy_data(client_reader, remote_writer),
                              loop=loop)

    tunnel_future = asyncio.ensure_future(
                asyncio.start_server(handle_client, *local_address, loop=loop),
                                          loop=loop)
    return I2PTunnel(session_name, writer, tunnel_future)

async def server_tunnel(local_address, loop=None, private_key=None, 
                    session_name=None, sam_address=i2plib.sam.DEFAULT_ADDRESS):
    """Run a server tunnel in the event loop.

    If you want to expose a local service 127.0.0.1:80 to the I2P network, run
    a server tunnel with a local address ("127.0.0.1", 80). If you don't 
    provide a private key or a session name, it will use a TRANSIENT 
    destination.

    :param local_address: A local address to bind a remote destination to. E.g.
                        ("127.0.0.1", 6668)
    :param session_name: (optional) Session nick name. A new session nickname is
                        generated if not specified.
    :param private_key: (optional) Private key to use in this session. Can be 
                        a base64 encoded string, i2plib.sam.PrivateKey instance
                        or None. TRANSIENT destination is used when it is None.
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :return: an instance of i2plib.tunnel.I2PTunnel
    """
    session_name = session_name or i2plib.sam.generate_session_id()
    reader, writer = await i2plib.aiosam.create_session(session_name,
                style="STREAM", sam_address=sam_address, loop=loop,
                private_key=private_key)

    async def handle_client(incoming, client_reader, client_writer):
        # data and dest may come in one chunk
        dest, data = incoming.split(b"\n", 1) 
        remote_destination = i2plib.sam.Destination(dest.decode())
        logging.debug("{} client connected: {}.b32.i2p".format(session_name,
            remote_destination.base32))

        try:
            remote_reader, remote_writer = await asyncio.wait_for(
                    asyncio.open_connection(
                       host=local_address[0], port=local_address[1], loop=loop),
                    timeout=5, loop=loop)
            if data: remote_writer.write(data)
            asyncio.ensure_future(proxy_data(remote_reader, client_writer),
                                  loop=loop)
            asyncio.ensure_future(proxy_data(client_reader, remote_writer),
                                  loop=loop)
        except ConnectionRefusedError:
            client_writer.close()

    async def server_loop():
        while True:
            client_reader, client_writer = await i2plib.aiosam.stream_accept(
                    session_name, sam_address=sam_address, loop=loop)
            incoming = await client_reader.read(BUFFER_SIZE)
            asyncio.ensure_future(handle_client(
                incoming, client_reader, client_writer), loop=loop)

    tunnel_future = asyncio.ensure_future(server_loop(), loop=loop)
    return I2PTunnel(session_name, writer, tunnel_future)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('type', metavar="TYPE", 
                        help="Tunnel type (server or client)")
    parser.add_argument('address', metavar="ADDRESS", 
                        help="Local address (e.g. 127.0.0.1:8000)")
    parser.add_argument('--debug', '-d', action='store_true',
                        help='Debugging')
    parser.add_argument('--key', '-k', default='', metavar='PRIVATE_KEY',
                        help='Path to private key file')
    parser.add_argument('--destination', '-D', default='', 
                        metavar='DESTINATION', help='Remote destination')
    args = parser.parse_args()

    SAM_ADDRESS = i2plib.utils.get_sam_address()

    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)
    loop = asyncio.get_event_loop()
    loop.set_debug(args.debug)

    if args.key:
        private_key = i2plib.sam.PrivateKey(path=args.key)
    else:
        private_key = None

    local_address = i2plib.utils.address_from_string(args.address)

    if args.type == 'client':
        asyncio.ensure_future(client_tunnel(local_address, 
            args.destination, loop=loop, private_key=private_key, 
            sam_address=SAM_ADDRESS), loop=loop)
    elif args.type == 'server':
        asyncio.ensure_future(server_tunnel(local_address, loop=loop,
            private_key=private_key, sam_address=SAM_ADDRESS), loop=loop)

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        loop.stop()
        loop.close()



import asyncio
import logging

import i2plib.sam
import i2plib.exceptions
import i2plib.utils

BUFFER_SIZE = 65536

def parse_reply(data):
    return i2plib.sam.Answer(data.decode().strip())

async def get_sam_socket(sam_address=i2plib.sam.DEFAULT_ADDRESS, loop=None):
    """A couroutine used to create a new SAM socket.

    :param sam_address: (optional) SAM API address
    :param loop: (optional) event loop instance
    :return: A (reader, writer) pair
    """
    reader, writer = await asyncio.open_connection(*sam_address, loop=loop)
    writer.write(i2plib.sam.hello("3.1", "3.1"))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        return (reader, writer)
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()

async def dest_lookup(domain, sam_address=i2plib.sam.DEFAULT_ADDRESS, 
                      loop=None):
    """A coroutine used to lookup a full I2P destination by .i2p domain or 
    .b32.i2p address.

    :param domain: Address to be resolved, can be a .i2p domain or a .b32.i2p 
                   address.
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :return: An instance of i2plib.Destination
    """
    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.naming_lookup(domain))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        writer.close()
        return i2plib.sam.Destination(reply["VALUE"])
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()

async def new_private_key(sam_address=i2plib.sam.DEFAULT_ADDRESS, loop=None,
                      sig_type=i2plib.sam.Destination.default_sig_type):
    """A coroutine used to generate a new private key of a chosen signature
    type.

    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :param sig_type: (optional) Signature type
    :return: An instance of i2plib.PrivateKey
    """
    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.dest_generate(sig_type))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    writer.close()
    return i2plib.sam.PrivateKey(reply["PRIV"])

async def create_session(session_name, sam_address=i2plib.sam.DEFAULT_ADDRESS, 
                         loop=None, session_ready=None, style="STREAM",
                         signature_type=i2plib.sam.Destination.default_sig_type,
                         private_key=None, options={}, session_created=None,
                         args=()):
    """A coroutine used to create a new SAM session.

    :param session_name: Session nick name
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :param session_ready: (optional) asyncio.Event instance to notify when 
                        session is ready 
    :param style: (optional) Session style, can be STREAM, DATAGRAM, RAW
    :param signature_type: (optional) If the destination is TRANSIENT, this 
                        signature type is used
    :param private_key: (optional) Private key to use in this session. Can be 
                        a base64 encoded string, i2plib.sam.PrivateKey instance
                        or None. TRANSIENT destination is used when it is None.
    :param options: (optional) A dict object with i2cp options
    :param session_created: (optional) A coroutine to be executed after the 
                        session is created. Executed with arguments
                        (loop, reader, writer, private_key, \*args)
    :param args: (optional) Arguments for a session_created coroutine
    :return: A (reader, writer) pair
    """
    logging.debug("Creating session {}".format(session_name))
    if private_key:
        if type(private_key) == i2plib.sam.PrivateKey:
            private_key = private_key
        else:
            private_key = i2plib.sam.PrivateKey(private_key)
        destination = private_key.base64
    else:
        private_key = None
        destination = i2plib.sam.TRANSIENT_DESTINATION

    options = " ".join(["{}={}".format(k, v) for k, v in options.items()])

    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.session_create(
            style, session_name, destination, options))

    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        if not private_key:
            private_key = i2plib.sam.PrivateKey(reply["DESTINATION"]) 
        logging.debug(private_key.destination.base32)
        if session_ready:
            session_ready.set()
        logging.debug("Session created {}".format(session_name))
        if session_created:
            asyncio.ensure_future(session_created(loop, reader, writer, 
                private_key, *args), loop=loop)
        return (reader, writer)
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()

async def stream_connect(session_name, destination, 
                         sam_address=i2plib.sam.DEFAULT_ADDRESS, loop=None, 
                         session_ready=None, stream_connected=None, args=()):
    """A coroutine used to connect to a remote I2P destination.

    :param session_name: Session nick name
    :param destination: I2P destination to connect to
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :param session_ready: (optional) asyncio.Event instance to notify when 
                        session is ready 
    :param stream_connected: (optional) A coroutine to be executed after the 
                        connection has been established. Executed with arguments
                        (loop, session_name, reader, writer, \*args)
    :param args: (optional) Arguments for a stream_connected coroutine
    :return: A (reader, writer) pair
    """
    logging.debug("Connecting stream {}".format(session_name))
    if isinstance(destination, str) and not destination.endswith(".i2p"):
        destination = i2plib.sam.Destination(destination)
    elif isinstance(destination, str):
        destination = await dest_lookup(destination, sam_address, loop)

    if session_ready: await session_ready.wait()
    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.stream_connect(session_name, destination.base64,
                                           silent="false"))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        logging.debug("Stream connected {}".format(session_name))
        if stream_connected:
            asyncio.ensure_future(
              stream_connected(loop, session_name, reader, writer, *args),
              loop=loop)
        return (reader, writer)
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()

async def stream_accept(session_name, sam_address=i2plib.sam.DEFAULT_ADDRESS,
                        loop=None, session_ready=None, stream_connected=None, 
                        args=()):
    """A coroutine used to accept a connection from the I2P network.

    :param session_name: Session nick name
    :param sam_address: (optional) SAM API address
    :param loop: (optional) Event loop instance
    :param session_ready: (optional) asyncio.Event instance to notify when 
                        session is ready 
    :param stream_connected: (optional) A coroutine to be executed after the 
                        connection has been established. Executed with arguments
                        (loop, session_name, reader, writer, \*args)
    :param args: (optional) Arguments for a stream_connected coroutine
    :return: A (reader, writer) pair
    """
    if session_ready: await session_ready.wait()
    reader, writer = await get_sam_socket(sam_address, loop)
    writer.write(i2plib.sam.stream_accept(session_name, silent="false"))
    reply = parse_reply(await reader.read(BUFFER_SIZE))
    if reply.ok:
        if stream_connected:
            asyncio.ensure_future(
              stream_connected(loop, session_name, reader, writer, *args), 
              loop=loop)
        return (reader, writer)
    else:
        raise i2plib.exceptions.SAM_EXCEPTIONS[reply["RESULT"]]()
